/*
 * usage: netchecker [-1DLPcdeginqrsv] [-E excludefilename] [-o outfile]
 *                   [-p probehost] [-R time ] [-t timer] conffilename
 *   -1      run once then quit (multiple -1 to run again on errors)
 *   -D      dump netstat output as-is
 *   -c      display config and exit
 *   -d      add debugging information
 *   -e      detect established sockets that are not in configuration file,
 *           and not in excluded list (if -E is used)
 *   -g      (guess dot) both : and . are considered as dot
 *   -i      output to be piped to itrs software
 *   -L      use lsof(1) instead of netstat(1)
 *   -n      numerical addresses for host, port or user names (netstat -n)
 *   -o      send output here instead of to stdout
 *   -P      use cat (or dd) /proc/net/tcp on Linux instead of lsof(1) or netstat(1)
 *   -p      host to send probe to
 *   -q      quiet, do not display if level=0
 *   -R      record until time
 *   -r      use regex in host and port
 *   -s      use syslog
 *   -t      rescan every timer second (default 30), if -1 not set
 *   -v      print version and exit
 * conffilename the name of the config file
 *
 * The safe way to terminate the program is SIGINT (ctrl-C) or SIGTERM
 * This utility was written by Philippe Charnier <charnier@xp11.frmug.org>
 */

/* the date is accurate but the revision number is not */
static const char rcsid[] =
  "$Id: netchecker.c,v 1.76 2011/12/16 09:32:08 phcha Exp $";

#include <arpa/inet.h>
#if defined(__FreeBSD__)
#include <sys/param.h>
#endif
#if defined(__linux__)
#include <asm/param.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netdb.h>
#include <netinet/in.h>
#include <ctype.h>
#include <libgen.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#if defined(__linux__)
#include <time.h>
#endif
#include <regex.h>
#include <unistd.h>

#define MAXBUFFERSIZE	4096*3	/* size of netstat(1)/lsof(1) output */
#define MAXLASTSOCKET	1024*2	/* # of sockets */
#define MAXCONFIGLIST	256	/* # of rules */
#define MAXEXCLUDELIST	256	/* # of excluded sockets */
#define MAXPROBELIST	4

#define STATEUNKNOWN	0
#define STATERUNNING	1
#define STATEWRONGTIME	2
#define STATEEXCLUDED	3

#define HPS    '/'     /* host port separator,
                        * must be != from ',' because of itrs(1)
                        * must be != from '*'
                        */
int loop;
int lastsocket;
int lastconfig;
int lastprobehost;
int lastexclude;
int time_hour_min;
int time_hour_min_sec;
int weekday;
char *category;
volatile int sig_terminate;
int timer;
char *conffilename;
char *excludefilename;
char hostname[MAXHOSTNAMELEN];
int record;
int sockets, socketsallowed;

FILE *outfile;

struct sprobehost {
  char *probehost;
  int   sockfd;
  struct sockaddr_in name;
  int   public; /* -1 for global probehost (#probehost ... or -p probehost) */
};

#define NETSTATMODE 0
#define LSOFMODE    1
#define CATPROCMODE 2

struct sflags {
  int debug;
  int dump_netstat;
  int excluded;
  int guessdot;
  int itrs;
  int callmode;
  int once;
  int regex;
  int syslog;
  int displayconfig;
  int quiet;
  int numeric; /* numerical addresses instead of trying to determine symbolic host, port or user names. */ 
};

struct stag {
  pid_t pid; /* used in lsof(1) mode or netstat(1)/Linux, 0 otherwise */
  char *saddr_sport;
  char *daddr_dport;
  int   start_hour_min;
  int   loop;
  int   match;
  int   state;
};

struct ctag {
  pid_t pid; /* obtained from pidfile */
  char *pidfile; /* pidfile = ... */
  char *saddr_sport;
  char *daddr_dport;
  int   start_hour_min;
  int   end_hour_min;
  int   alarm_level;
  int   weekday;
  int   match;
  int   state;
  char *comment;
  char *category;
  int   delay;
  int   regexmode;
};

struct etag {
  char *socket;
};

struct sflags f;
struct sprobehost probehostlist[MAXPROBELIST];
struct stag socketlist[MAXLASTSOCKET];
struct ctag configlist[MAXCONFIGLIST];
struct etag excludelist[MAXEXCLUDELIST];

/*
 * display error message and exit
 */
void
myerrx(int exitcode, char *msg)
{
  fprintf(stderr, "netchecker: %s\n", msg);
  exit(exitcode);
}

/*
 * remove trailing spaces and \n if any
 */
void
chomp(char *buffer)
{ char *ptr, *last_not_space;

  ptr = last_not_space = buffer;
  while (*ptr) {
    if (!isspace((int)*ptr))
      last_not_space = ptr;
    ptr++;
  }
  if (*last_not_space && !isspace((int)*last_not_space))
    last_not_space++;
  *last_not_space = '\0';
}

/*
 * send an UDP datagram to distant host
 */
void
sendprobemessage(char *msg, int public)
{
  int nmsg, i;

  for (i = 0; i < lastprobehost; i++) {
    if (public == -2 ||
        probehostlist[i].public == public ||
        probehostlist[i].public == -1) {
      nmsg = strlen(msg);
      if (sendto(probehostlist[i].sockfd, msg, nmsg, 0,
          (struct sockaddr *)&(probehostlist[i].name),
          sizeof(probehostlist[i].name)) != nmsg)
        myerrx(1, "sendto");
    }
  }
}

/*
 * locate the first non space in the string
 */
void
skipwhitespace(char **ptr)
{
  while(*ptr && isspace((int)**ptr)) (*ptr)++;
}

/*
 * locate the "->" substring and replace it with "  ". Used in lsof(1) mode.
 */
void
skiparrow(char *ptr)
{
  while(*ptr) {
   if (strncmp(ptr, "->", sizeof("->") - 1) == 0) {
     *ptr = ' ';
     ptr++;
     *ptr = ' ';
   }
   ptr++;
 }
}

int
incomment(char *ptr)
{ static char *start = NULL;
  static char *stop  = NULL;
  int state;

  if (ptr == NULL) {
    start = stop = NULL;
    return (0);
  }

  if (start == NULL) {
    start = ptr;
    while (*start && strncmp("comment=", start, sizeof("comment=") - 1) != 0)
      start++;
    stop = start;
    if (*start) {
      start += sizeof("comment=") - 1;
      stop  = start + 1;
      state = 0;
      while (*stop && state != 2) {
        switch (state) {
          case 0: if (*stop == '\\') {
                    stop++;
                    state = 1; /* escape next char */
                  }
                  else if (*stop == *start) { state = 2; /* finish */}
                  else { stop++; }
                  break;
          case 1: stop++;
                  state = 0;
                  break;
        }
      }
    }
  }

  return (start < ptr && ptr < stop);
}

/*
 * get the next field in a space(s) separated fields of a string
 */
char *
shift (char **remain, int stopatspace)
{ char *result;

  while(**remain && isspace((int)**remain)) (*remain)++;
  result = *remain;
  if (4 <= f.debug)
    fprintf(outfile, "shift(): ##%s##\n", result);
  if (stopatspace) {
    incomment(NULL); /* initialise */
    while(**remain && !(!incomment(*remain) &&
          isspace((int)**remain)))
      (*remain)++;
    if (!incomment(*remain) && isspace((int)**remain)) {
      **remain = '\0';
      (*remain)++;
    }
  }
  else {
    while(**remain && **remain != '\n') (*remain)++;
    **remain = '\0';
  }
  return (result);
}

/*
 * search for a matching regex
 */
int
matchingextended(char *rgx, char *str)
{ int rv;
  regex_t reg;

  rv = regcomp(&reg, rgx, REG_EXTENDED|REG_NOSUB);
  if (rv != 0) {
    if (4 <= f.debug)
      fprintf(outfile, "matchingextended()/regcomp(): ##%s## and ##%s## returned %d\n", rgx, str, rv);
    myerrx(1, "regcomp");
  }
  rv = regexec(&reg, str, 0, NULL, 0);
  regfree(&reg);
  if (4 <= f.debug)
    fprintf(outfile, "matchingextended(): regex=##%s## and string=##%s## %s\n", rgx, str, (rv == 0) ? "[match]" : "[no match]");
  return (rv == 0);
}

/*
 * search for a matching regex
 */
int
isexcludedextended(char *saddr_sport, char *daddr_dport)
{ int i;
  char buffer[1024];

  sprintf(buffer, "%s=%s", saddr_sport, daddr_dport);

  for (i = 0; i < lastexclude; i++)
    if (matchingextended(excludelist[i].socket, buffer))
      return (1);
  return (0);
}

/*
 * ouput the contents
 */
void
dumpall(void)
{ int i;

  if (!(2 <= f.debug)) return;

  for (i = 0; i < lastsocket; i++)
    fprintf(outfile, "  %04d 2400 %ld %s %s\n", socketlist[i].start_hour_min,
            (long)socketlist[i].pid, socketlist[i].saddr_sport, socketlist[i].daddr_dport);
}

/*
 * output an itrs software compatible report
 */
void
dumpconfigtoitrs(void)
{ int i, j;
  char *levelstr, *statusstr;

  levelstr = "(null)";
  statusstr = "(null)";
  printf ("index,comment,level,status,category,pid,src_addr/port,dest_addr/port,"
          "start time,end time..\n");
  for (i = 0; i < lastconfig; i++) {
    if ((configlist[i].weekday & weekday) == 0)
      continue;
    switch(configlist[i].state) {
      case STATEUNKNOWN : /* still unknown, may be missing */
                          if (time_hour_min < configlist[i].start_hour_min
                              || configlist[i].end_hour_min <= time_hour_min) {
                            levelstr = "OK";
                            statusstr = "not checked";
                            break;
                          }

                          if (50 < configlist[i].alarm_level)
                            levelstr = "CRITICAL";
                          else if (20 < configlist[i].alarm_level)
                            levelstr = "WARNING";
                          else
                            levelstr = "STANDBY";
                          statusstr = "missing";
                          if (configlist[i].alarm_level == 0) {
                            levelstr = "OK";
                            statusstr = "not checked";
                          }
                          break;
      case STATERUNNING :
                          levelstr = "OK";
                          statusstr = "running";
                          break;
      case STATEWRONGTIME :
                          levelstr = "WARNING";
                          statusstr = "wrong time";
                          break;
    }
    if (configlist[i].alarm_level != 0 || f.quiet == 0)
      printf ("%d,%s,%s,%s,%s,%ld,%s,%s,%04d,%04d\n", i,
            configlist[i].comment, levelstr, statusstr,
            configlist[i].category,
            (long)((configlist[i].match != -1) ? socketlist[configlist[i].match].pid : 0), configlist[i].saddr_sport,
            configlist[i].daddr_dport, configlist[i].start_hour_min,
            configlist[i].end_hour_min);
  }
  /* Now, display unknown established sockets. */
  if (f.excluded) {
    for (j = 0; j < lastsocket; j++) {
      if (socketlist[j].state == STATEUNKNOWN) {
        printf ("%d,%s,%s,%s,%s,%ld,%s,%s,%04d,%04d\n", i, "n/a", "OK",
                "unknown", "n/a", (long)socketlist[j].pid,
                socketlist[j].saddr_sport,
                socketlist[j].daddr_dport, socketlist[j].start_hour_min, 0);
        i++;
      }
    }
  }
}

/*
 * strings are in the form <hostname>,<port> and we assume that `,' is not
 * present elsewhere. Find if the two strings match. The substring "*" match
 * one or more standard characters (no , and no \0).
 */
int
matching(char *rgx, char *str, int regexmode)
{
  if (4 <= f.debug)
    fprintf(outfile, "matching(): ##%s## and ##%s## under regexmode = %d\n", rgx, str, regexmode);

  if (regexmode)
    return (matchingextended(rgx, str));

  while (1) {
    switch(*rgx) {
      case '*': if (*str == HPS || *str == '\0') rgx++;
                else {
                  if (matching(rgx+1, str+1, regexmode)) {
                    if (4 <= f.debug)
                      fprintf(outfile, "matching(): regex=##%s## and string=##%s## %s\n", rgx, str, "[match]");
                    return (1);
                  }
                  else
                    str++;
                }
                break;
      case HPS: if (*str != HPS) {
                  if (4 <= f.debug)
                    fprintf(outfile, "matching(): regex=##%s## and string=##%s## %s\n", rgx, str, "[no match]");
                  return (0);
                }
                rgx++; str++;
                break;
      case '\0':return (*str == '\0');
      default:  if (*str == HPS || *str == '\0' || *rgx != *str) {
                  if (4 <= f.debug)
                    fprintf(outfile, "matching(): regex=##%s## and string=##%s## %s\n", rgx, str, "[no match]");
                  return (0);
                }
                rgx++; str++;
                break;
    }
  }
}

/*
 * convert a binary string "0111100" representing a mask of the days in a
 * week to an integer
 */
int
atobin(char *str)
{ int result = 0;
  char *ptr;

  if (strlen(str) != 7)
    myerrx(1, "incorrect weekday length");

  ptr = str;
  while(*ptr && (*ptr == '1' || *ptr == '0'))
    ptr++;

  if (*ptr)
    myerrx(1, "incorrect weekday definition");

  ptr = str;
  while (*ptr) {
    result *= 2;
    result += (*ptr == '1') ? 1 : 0;
    ptr++;
  }
  return (result);
}

/*
 * parse config file for directives, comments, blank lines
 */
void readexcludeentry(char *);

int
commentorblank(char *buffer)
{ char *from, *to;
  int state;

#define READDIRECT(x) if (strncmp(buffer, x, sizeof(x) - 1) == 0)

  chomp(buffer);

  READDIRECT("#excluded_flag 0") {
    f.excluded = 0;
  }
  READDIRECT("#excluded_flag off") {
    f.excluded = 0;
  }
  READDIRECT("#excluded_flag 1") {
    f.excluded = 1;
  }
  READDIRECT("#excluded_flag on") {
    f.excluded = 1;
  }

  READDIRECT("#quiet_flag 0") {
    f.quiet = 0;
  }
  READDIRECT("#quiet_flag off") {
    f.quiet = 0;
  }
  READDIRECT("#quiet_flag 1") {
    f.quiet = 1;
  }
  READDIRECT("#quiet_flag on") {
    f.quiet = 1;
  }

  READDIRECT("#numeric_flag 0") {
    f.numeric = 0;
  }
  READDIRECT("#numeric_flag off") {
    f.numeric = 0;
  }
  READDIRECT("#numeric_flag 1") {
    f.numeric = 1;
  }
  READDIRECT("#numeric_flag on") {
    f.numeric = 1;
  }

  READDIRECT("#lsofmode_flag 0") {
    f.callmode = NETSTATMODE;
  }
  READDIRECT("#lsofmode_flag off") {
    f.callmode = NETSTATMODE;
  }
  READDIRECT("#lsofmode_flag 1") {
    f.callmode = LSOFMODE;
  }
  READDIRECT("#lsofmode_flag on") {
    f.callmode = LSOFMODE;
  }

  READDIRECT("#catprocmode_flag 0") {
    f.callmode = NETSTATMODE;
  }
  READDIRECT("#catprocmode_flag off") {
    f.callmode = NETSTATMODE;
  }
  READDIRECT("#catprocmode_flag 1") {
    f.callmode = CATPROCMODE;
  }
  READDIRECT("#catprocmode_flag on") {
    f.callmode = CATPROCMODE;
  }

  READDIRECT("#regex_flag 0") {
    f.regex = 0;
  }
  READDIRECT("#regex_flag off") {
    f.regex = 0;
  }
  READDIRECT("#regex_flag 1") {
    f.regex = 1;
  }
  READDIRECT("#regex_flag on") {
    f.regex = 1;
  }

  READDIRECT("#excluded_rule") {
    readexcludeentry(&buffer[sizeof("#excluded_rule")]);
  }

  READDIRECT("#probehost") {
    /*
     * e.g. #probehost host_where_servchecker_is_listening
     * trivial syntax required: one space before hostname, no space after
     */
    if (lastprobehost == MAXPROBELIST)
      myerrx(1, "increase MAXPROBELIST");
    probehostlist[lastprobehost].public = -1;
    probehostlist[lastprobehost].probehost =
                                 strdup(&buffer[sizeof("#probehost")]);
    lastprobehost++;
  }

  READDIRECT("#weekday") {
    weekday = atobin(&buffer[sizeof("#weekday")]);
  }

  READDIRECT("#sockets") {
    socketsallowed = atoi(&buffer[sizeof("#sockets")]);
  }

  READDIRECT("#category") {
    free(category);
    category = strdup(&buffer[sizeof("#category")]);
  }

#undef READDIRECT

  /*
   * if # is escaped, replace \# with # and continue. If # is not escaped
   * thus starting a comment, remove after it.
   */
  if (4 <= f.debug)
    fprintf(outfile, "commentorblank(): ##%s##", buffer);

  from = to = buffer;
  state = 0;
  while (*from) {
    switch (state) {
    case 0 : if (*from == '#') { *from = '\0'; /* finish */ }
             else if (*from == '\\') {
               from++;
               state = 1; /* escape next char */
             }
             else { *to = *from; to++; from++; }
             break;
    case 1 : if (*from != '#')
               *to++ = '\\';
             *to = *from; to++; from++;
             state = 0;
             break;
    }
  }
  *to = '\0';

  if (4 <= f.debug)
    fprintf(outfile, " replaced with ##%s##\n", buffer);

  return (*buffer == '\0');
}

/*
 * read line for excluded sockets
 */
void
readexcludeentry(char *buffer)
{ char *remainder;

  if (commentorblank(buffer))
    return;
  if (lastexclude == MAXEXCLUDELIST)
    myerrx(1, "increase MAXEXCLUDELIST");
  remainder = buffer;
  skipwhitespace(&remainder);
  excludelist[lastexclude].socket =
      strdup(shift(&remainder, 0 /* do not stop at space */));
  lastexclude++;
}

/*
 * read pid file
 */
void
readpidfile(int c)
{ char buffer[32];
  FILE *stream;

#if !defined(__linux__)
  if (f.callmode != LSOFMODE) return;
#endif

  if (c != -1) {
    stream = fopen(configlist[c].pidfile, "r");
    if (stream == NULL) {
      configlist[c].pid = 0;
      return;
    }
    if ((fgets(buffer, sizeof(buffer) - 1, stream)) != NULL)
      configlist[c].pid = atol(buffer);
    else
      configlist[c].pid = 0;
    fclose(stream);
    return;
  }
  else {
    int i;

    for (i = 0; i < lastconfig; i++)
      if (0 != strcmp(configlist[i].pidfile, "n/a")) readpidfile(i);
  }
}

/*
 * read file for excluded sockets
 */
void
readexcludefile(void)
{ char buffer[1024];
  FILE *stream;

  if (excludefilename == NULL)
    return;

  stream = fopen(excludefilename, "r");
  if (stream == NULL) {
    fprintf(stderr, "netchecker: can not read file %s\n", excludefilename);
    exit(1);
  }
  while ((fgets(buffer, sizeof(buffer) - 1, stream)) != NULL)
    readexcludeentry(buffer);
  fclose(stream);
}

char *
strdupcomment(char *ptr)
{ char *result;
  char *from, *to;
  int state;

  result = (char *)malloc(strlen(ptr) + 1);
  if (result == NULL)
    return (NULL);

  if (4 <= f.debug)
    fprintf(outfile, "strdupcomment(): ##%s##\n", ptr);

  from = ptr;
  to = result;

  from  = ptr + 1;
  state = 0;
  while (*from && state != 2) {
    switch (state) {
    case 0: if (*from == '\\') {
              from++;
              state = 1; /* escape next char */
            }
            else if (*from == *ptr) { state = 2; /* finish */}
            else { *to = *from; to++; from++; }
            break;
    case 1: *to = *from; to++; from++;
            state = 0;
            break;
    }
  }
  *to = '\0';
  return (result);
}

/*
 * substitute dot separator between hostname and port
 * hostname . port -> hostname ; port
 * Linux uses `:' instead of `.'
 * check that only `;' and `\0' can follow `*'
 */
void
normalize(char *addrdotport, int mode)
{ char *lastdot;
  char *ptr;
  char dotsign;

  if (4 <= f.debug)
    fprintf(outfile, "normalize(): ##%s##\n", addrdotport);

  switch (mode) {
    case LSOFMODE :    dotsign = ':';
                       break;
    case CATPROCMODE :
#if defined(__linux__)
                       dotsign = ':';
#else
                       dotsign = '.';
#endif
                       break;
    default :
#if defined(__linux__)
                       dotsign = ':';
#else
                       dotsign = '.';
#endif
  }

  lastdot = strrchr(addrdotport, dotsign);

  if (f.guessdot) {
    ptr = addrdotport;
    lastdot = NULL;

    while (*ptr != '\0') {
      if (*ptr == ':' || *ptr == '.')
        lastdot = ptr;
      ptr++;
    }
  }

  if (lastdot == NULL) {
    if (dotsign == ':')
      myerrx(1, "missing colon in <hostname>:<port>");
    else
      myerrx(1, "missing dot in <hostname>.<port>");
  }
  *lastdot = HPS;

  while (*addrdotport) {
    if (*addrdotport == '*') {
      addrdotport++;
      if (*addrdotport != HPS && *addrdotport) {
        if (dotsign == ':')
          myerrx(1, "badly placed * in <hostname>:<port>");
        else
          myerrx(1, "badly placed * in <hostname>.<port>");
      }
      continue;
    }
    addrdotport++;
  }
}

/*
 * display config file
 */
void
displayconfig(void)
{ int i;

  for (i = 0; i < lastconfig; i++) {
    /* need -cc to display if alarm_level is 0 */
    if (1 < f.displayconfig || configlist[i].alarm_level != 0)
      printf("%s=%s %04d %04d %d\n",
        configlist[i].saddr_sport, configlist[i].daddr_dport,
        configlist[i].start_hour_min, configlist[i].end_hour_min,
        configlist[i].alarm_level);
  }
}

/*
 * read config file
 */
void
readconffile(void)
{ char *remainder, *ptr, *ptr2, *result;
  char buffer[1024];
  int found;
  FILE *stream;
  int repeat;

  lastconfig = 0;

  /*
   * addresses and ports are obtained from netstat(1):
   * Solaris /usr/bin/netstat -P tcp
   * FreeBSD /usr/bin/netstat -p tcp
   * Linux   /bin/netstat -t
   * or lsof(1):
   * lsof -s -l -i TCP  (based on lsof -s -l -a -p 25900,23683 -i TCP)
   */
  stream = fopen(conffilename, "r");
  if (stream == NULL) {
    fprintf(stderr, "netchecker: can not read file %s\n", conffilename);
    exit(1);
  }
  while ((fgets(buffer, sizeof(buffer) - 1, stream)) != NULL) {
    if (commentorblank(buffer))
      continue;
    if (lastconfig == MAXCONFIGLIST)
      myerrx(1, "increase MAXCONFIGLIST");
    remainder = buffer;
    skipwhitespace(&remainder);
    configlist[lastconfig].pid = 0;
    configlist[lastconfig].pidfile = strdup("n/a");
    configlist[lastconfig].start_hour_min = atoi(shift(&remainder, 1 /* stop at first space */));
    configlist[lastconfig].end_hour_min = atoi(shift(&remainder, 1 /* stop at first space */));

    configlist[lastconfig].alarm_level = 90; /* default value */
    configlist[lastconfig].weekday = weekday; /* default value */
    configlist[lastconfig].comment = strdup("n/a");
    configlist[lastconfig].category = strdup(category); /* default value */
    configlist[lastconfig].delay = 0; /* default value */
    configlist[lastconfig].regexmode = f.regex; /* default value */
    ptr = ptr2 = strdup(shift(&remainder, 1 /* stop at first space */));
    repeat = 1;
    while ((result = strtok(ptr2, ","))) {
      ptr2 = NULL;
      found = 0;
      if (found == 0 && strncmp("level=", result, sizeof("level=") - 1) == 0) {
        configlist[lastconfig].alarm_level = atoi(&result[sizeof("level=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("weekday=", result, sizeof("weekday=") - 1) == 0) {
        configlist[lastconfig].weekday = atobin(&result[sizeof("weekday=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("pidfile=", result, sizeof("pidfile=") - 1) == 0) {
        free(configlist[lastconfig].pidfile);
        configlist[lastconfig].pidfile = strdup(&result[sizeof("pidfile=") - 1]);
        readpidfile(lastconfig);
        /* pidfile need to be reread when netstat(1) is run */
        found = 1;
      }
      if (found == 0 && strncmp("comment=", result, sizeof("comment=") - 1) == 0) {
        free(configlist[lastconfig].comment);
        configlist[lastconfig].comment = strdupcomment(&result[sizeof("comment=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("category=", result, sizeof("category=") - 1) == 0) {
        free(configlist[lastconfig].category);
        configlist[lastconfig].category = strdup(&result[sizeof("category=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("probehost=", result, sizeof("probehost=") - 1) == 0) {
        if (lastprobehost == MAXPROBELIST)
          myerrx(1, "increase MAXPROBELIST");
        probehostlist[lastprobehost].public = lastconfig;
        probehostlist[lastprobehost].probehost =
            strdup(&result[sizeof("probehost=") - 1]);
        lastprobehost++;
        found = 1;
      }
      if (found == 0 && strncmp("repeat=", result, sizeof("repeat=") - 1) == 0) {
        repeat = atoi(&result[sizeof("repeat=") - 1]);
        found = 1;
      }
      if (found == 0 && strncmp("delay=", result, sizeof("delay=") - 1) == 0) {
        configlist[lastconfig].delay = atoi(&result[sizeof("delay=") - 1]);
        found = 1;
      }
      if (found == 0) {
        if (1 <= f.debug)
            fprintf(outfile, "netchecker: unknown option field: %s\n", result);
        if (1 <= f.syslog)
            syslog(LOG_ERR, "netchecker: unknown option field: %s", result);
        /* do not abort here, to make old config work exit(1); */
      }
    }
    free(ptr);

    configlist[lastconfig].saddr_sport = strdup(shift(&remainder, 1 /* stop at first space */));
    normalize(configlist[lastconfig].saddr_sport, NETSTATMODE);
    configlist[lastconfig].daddr_dport = strdup(shift(&remainder, 1 /* stop at first space */));
    normalize(configlist[lastconfig].daddr_dport, NETSTATMODE);
    configlist[lastconfig].match = -1;
    configlist[lastconfig].state = STATEUNKNOWN;
    lastconfig++;
    while (1 < repeat) {
      if (lastconfig == MAXCONFIGLIST)
        myerrx(1, "increase MAXCONFIGLIST");
      configlist[lastconfig].pid = configlist[lastconfig-1].pid;
      configlist[lastconfig].pidfile = strdup(configlist[lastconfig-1].pidfile);
      configlist[lastconfig].start_hour_min = configlist[lastconfig-1].start_hour_min;
      configlist[lastconfig].end_hour_min = configlist[lastconfig-1].end_hour_min;
      configlist[lastconfig].alarm_level = configlist[lastconfig-1].alarm_level;
      configlist[lastconfig].weekday = configlist[lastconfig-1].weekday;
      configlist[lastconfig].saddr_sport = strdup(configlist[lastconfig-1].saddr_sport);
      configlist[lastconfig].daddr_dport = strdup(configlist[lastconfig-1].daddr_dport);
      configlist[lastconfig].comment = strdup(configlist[lastconfig-1].comment);
      configlist[lastconfig].category = strdup(configlist[lastconfig-1].category);
      configlist[lastconfig].match = configlist[lastconfig-1].match;
      configlist[lastconfig].state = configlist[lastconfig-1].state;
      configlist[lastconfig].delay = configlist[lastconfig-1].delay;
      configlist[lastconfig].regexmode = configlist[lastconfig-1].regexmode;
      repeat--;
      lastconfig++;
    }
  }
  fclose(stream);
}

/*
 * reset data structures
 */
void
resetstate(void)
{ int i;

  for (i = 0; i < lastsocket; i++)
    socketlist[i].state = STATEUNKNOWN;
  for (i = 0; i < lastconfig; i++)
    configlist[i].state = STATEUNKNOWN;
}

int
matchpids(int c, int i)
{
  if (configlist[c].pid == 0 || socketlist[i].pid == 0) return (1);
  return (configlist[c].pid == socketlist[i].pid);
}

/*
 * create a matching relation between configlist[] and socketlist[]
 */
int
checkmatching(int c, int i)
{ int ii;

  if ((configlist[c].weekday & weekday) == 0)
    return (0);

  if (3 <= f.debug) {
    for (ii = 0; ii < lastsocket; ii++)
      if (socketlist[ii].match != -1 &&
          configlist[socketlist[ii].match].match != ii)
        fprintf(outfile, "checkmatching(): failed\n");
    for (ii = 0; ii < lastconfig; ii++)
      if (configlist[ii].match != -1 &&
          socketlist[configlist[ii].match].match != ii)
        fprintf(outfile, "checkmatching(): failed\n");
  }

  if (configlist[c].match == i && socketlist[i].match == c)
    return (1);

  if (matching(configlist[c].saddr_sport, socketlist[i].saddr_sport, configlist[c].regexmode) &&
      matching(configlist[c].daddr_dport, socketlist[i].daddr_dport, configlist[c].regexmode) &&
      matchpids(c, i)) {
    configlist[c].match = i;
    socketlist[i].match = c;
    return (1);
  }
  return (0);
}

/*
 * find a match
 */
int
isrunning(int c, int newstate)
{ int i;

  for (i = 0; i < lastsocket; i++) {
    if (socketlist[i].state == STATEUNKNOWN && checkmatching(c, i)) {
      socketlist[i].state = newstate;
      configlist[c].state = newstate;
      return (1);
    }
  }
  return (0);
}

/*
 * hide excluded sockets
 */
void
markexcluded(void)
{ int i;

  for (i = 0; i < lastsocket; i++)
    if (socketlist[i].state == STATEUNKNOWN &&
        isexcludedextended(socketlist[i].saddr_sport, socketlist[i].daddr_dport))
      socketlist[i].state = STATEEXCLUDED;
}

/*
 * show left sockets
 */
void
unknown(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastsocket; i++) {
    if (socketlist[i].state == STATEUNKNOWN) {
      if (1 <= f.debug)
        fprintf(outfile, "unknown established socket: %ld %s %s\n",
            (long)socketlist[i].pid, socketlist[i].saddr_sport, socketlist[i].daddr_dport);
      else {
        sprintf(msg, ":h=%s:a=N-UNKNOWN_RUNNING:l=%d:pid=%ld:x=%s %s:", hostname,
            timer,
            (long)socketlist[i].pid, socketlist[i].saddr_sport, socketlist[i].daddr_dport);
        sendprobemessage(msg, -1);
      }
      if (1 <= f.syslog)
        syslog(LOG_ERR, "unknown established socket: %ld %s %s",
            (long)socketlist[i].pid, socketlist[i].saddr_sport, socketlist[i].daddr_dport);
    }
  }
}

int
addtime(int c)
{ int minutes, hours;

  minutes = (configlist[c].start_hour_min % 100) + configlist[c].delay;
  hours   = configlist[c].start_hour_min / 100;

  while (60 <= minutes) {
    hours++;
    minutes-=60;
  }

  return ((24 <= hours) ?  2400 : hours * 100 + minutes);
}

/*
 * show missing sockets
 */
int
missing(void)
{ int i;
  char msg[1024];
  int e;

  e = 0;
  for (i = 0; i < lastconfig; i++) {
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (addtime(i) <= time_hour_min &&
        time_hour_min < configlist[i].end_hour_min) {
      /* sockets should be established */
      if (!isrunning(i, STATERUNNING)) {
        e++;
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile, "missing at %06d: %s %s {%04d %04d %d}\n",
                    time_hour_min_sec,
                    configlist[i].saddr_sport, configlist[i].daddr_dport,
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    configlist[i].alarm_level);
	}
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
	  if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR, "missing at %06d: %s %s {%04d %04d %d}",
                     time_hour_min_sec,
                     configlist[i].saddr_sport, configlist[i].daddr_dport,
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     configlist[i].alarm_level);
            if (90 < configlist[i].alarm_level)
              sprintf(msg, ":h=%s:a=N-MISSING-URGENT:l=%d:c=%s:x=%s %s:", hostname,
                    timer,
                    configlist[i].category, configlist[i].saddr_sport, configlist[i].daddr_dport);
            else if (20 < configlist[i].alarm_level)
              sprintf(msg, ":h=%s:a=N-MISSING:l=%d:c=%s:x=%s %s:", hostname,
                    timer,
                    configlist[i].category, configlist[i].saddr_sport, configlist[i].daddr_dport);
            else
              sprintf(msg, ":h=%s:a=N-MISSING-STANDBY:l=%d:c=%s:x=%s %s:", hostname,
                    timer,
                    configlist[i].category, configlist[i].saddr_sport, configlist[i].daddr_dport);
            sendprobemessage(msg, i);
          }
        }
      }
    }
  }
  return (e);
}

/*
 * show established sockets during wrong time
 */
void
running(void)
{ int i;
  char msg[1024];

  for (i = 0; i < lastconfig; i++) {
    if ((configlist[i].weekday & weekday) == 0)
      continue;

    if (time_hour_min < configlist[i].start_hour_min ||
        configlist[i].end_hour_min < time_hour_min) {
      /* sockets should not be established */
      if (isrunning(i, STATEWRONGTIME)) {
        if (1 <= f.debug) {
          if (configlist[i].alarm_level != 0)
            fprintf(outfile,
                    "established outside of [%04d,%04d] at %06d: %s %s\n",
                    configlist[i].start_hour_min, configlist[i].end_hour_min,
                    time_hour_min_sec,
                    configlist[i].saddr_sport, configlist[i].daddr_dport);
        }
        else {
          /*
           * use time_hour_min_sec to be sure that all messages will be
           * differents. They will not be buffered by a distant syslogd
           */
          if (configlist[i].alarm_level != 0) {
            if (1 <= f.syslog)
              syslog(LOG_ERR,
                     "established outside of [%04d,%04d] at %06d: %s %s",
                     configlist[i].start_hour_min, configlist[i].end_hour_min,
                     time_hour_min_sec,
                     configlist[i].saddr_sport, configlist[i].daddr_dport);
            sprintf(msg, ":h=%s:a=N-OUTSIDE_OF:l=%d:c=%s:x=%s %s:", hostname,
                    timer,
                    configlist[i].category, configlist[i].saddr_sport, configlist[i].daddr_dport);
            sendprobemessage(msg, i);
          }
        }
      }
    }
  }
}

/*
 * remove established sockets left from a previous loop
 */
void
removeoldconnections(void)
{ int i;

  i = 0;
  while (i < lastsocket) {
    if (socketlist[i].loop != loop) {
      if (2 <= f.debug || time_hour_min < record)
        fprintf(outfile, "- %04d %04d %ld %s %s\n",
                socketlist[i].start_hour_min, time_hour_min,
                (long)socketlist[i].pid, socketlist[i].saddr_sport, socketlist[i].daddr_dport);
      free(socketlist[i].saddr_sport);
      free(socketlist[i].daddr_dport);
      if (socketlist[i].match != -1) {
        configlist[socketlist[i].match].match = -1;
        socketlist[i].match = -1;
      }
      lastsocket--;
      if (i != lastsocket) {
        socketlist[i].pid = socketlist[lastsocket].pid;
        socketlist[i].saddr_sport = socketlist[lastsocket].saddr_sport;
        socketlist[i].daddr_dport = socketlist[lastsocket].daddr_dport;
        socketlist[i].loop= socketlist[lastsocket].loop;
        socketlist[i].start_hour_min = socketlist[lastsocket].start_hour_min;
        socketlist[i].match = socketlist[lastsocket].match;
        if (socketlist[i].match != -1) {
          configlist[socketlist[i].match].match = i;
        }
        socketlist[i].state = socketlist[lastsocket].state;
        i--;
      }
    }
    i++;
  }
}

char *
proc2ip(char *addrport, char *resultaddrport)
{ uint32_t result, addr;
  struct hostent *hp;
  struct servent *sp;
  char *ptr;
  unsigned char *xx;

  /*printf("enter proc2ip(): %s\n", addrport);*/

  /* 0147880A:1FA5 -> 10.136.71.1:8101 -> papps001:8101 */

  ptr = addrport;
  result = 0;
  while (*ptr != ':' && *ptr != '\0') {
    if ('0' <= *ptr && *ptr <= '9') { result <<= 4; result += *ptr - '0'; }
    if ('a' <= *ptr && *ptr <= 'f') { result <<= 4; result += *ptr - 'a' + 10; }
    if ('A' <= *ptr && *ptr <= 'F') { result <<= 4; result += *ptr - 'A' + 10; }
    ptr++;
  }
  *ptr = '\0';
  addr = result;
  xx = (char *)&addr;
  /*
  printf("*debug* trying %d.%d.%d.%d\n", xx[0], xx[1], xx[2], xx[3]);
  printf("*debug* %x\n", addr);
  */

  hp = gethostbyaddr(xx, 4, AF_INET);

  ptr++; /* skip NUL */

  result = 0;
  while (*ptr != '\0') {
    if ('0' <= *ptr && *ptr <= '9') { result <<= 4; result += *ptr - '0'; }
    if ('a' <= *ptr && *ptr <= 'f') { result <<= 4; result += *ptr - 'a' + 10; }
    if ('A' <= *ptr && *ptr <= 'F') { result <<= 4; result += *ptr - 'A' + 10; }
    ptr++;
  }

  sp = getservbyport(htons(result), "tcp");

  if (sp != NULL) {
    if (hp != NULL)
      sprintf(resultaddrport, "%s:%s", hp->h_name, sp->s_name);
    else
      sprintf(resultaddrport, "%s:%s", addrport, sp->s_name);
  }
  else {
    if (hp != NULL)
      sprintf(resultaddrport, "%s:%d", hp->h_name, result);
    else
      sprintf(resultaddrport, "%s:%d", addrport, result);
  }
  return (resultaddrport);
}

/*
 * compute a line of netstat(1)/lsof(1) output
 */
void
processline(char *buffer)
{ pid_t pid;
  char *saddr_sport, *daddr_dport;
  char resultsaddr_sport[256], resultdaddr_dport[256];
  char *State;
  char *remainder;
  int i;

  chomp(buffer);

  if (4 <= f.debug || 1 <= f.dump_netstat)
    fprintf (outfile, "processline(): ##%s##\n", buffer);

  remainder = buffer;
  if (LSOFMODE == f.callmode)
    skiparrow(buffer);
  skipwhitespace(&remainder);
  switch (f.callmode) {
    case LSOFMODE : /* process */ shift(&remainder, 1 /* stop at first space */);
                    pid = atol(shift(&remainder, 1 /* stop at first space */));
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() pid: ##%ld##\n", (long)pid);
                    /* user */ shift(&remainder, 1 /* stop at first space */);
                    /* fd */ shift(&remainder, 1 /* stop at first space */);
                    /* type */ shift(&remainder, 1 /* stop at first space */);
                    /* device */ shift(&remainder, 1 /* stop at first space */);
#if !defined(__linux__)
                    /* size */ shift(&remainder, 1 /* stop at first space */);
#endif
                    /* node */ shift(&remainder, 1 /* stop at first space */);
                    saddr_sport = shift(&remainder, 1 /* stop at first space */);
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() saddr_sport: ##%s##\n", saddr_sport);
                    daddr_dport = shift(&remainder, 1 /* stop at first space */);
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() daddr_dport: ##%s##\n",  daddr_dport);
                    State = shift(&remainder, 0 /* do not stop at space */);
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() State: ##%s##\n", State);
                    break;
    case CATPROCMODE :
                    pid = 0;
                    /* sl */ shift(&remainder, 1 /* stop at first space */);
                    saddr_sport = shift(&remainder, 1 /* stop at first space */);
                    if (0 == strcmp(saddr_sport, "records")) return;
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() saddr_sport: ##%s##\n", saddr_sport);
                    daddr_dport = shift(&remainder, 1 /* stop at first space */);
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() daddr_dport: ##%s##\n",  daddr_dport);
                    State = shift(&remainder, 1 /* stop at space */);
                    if (4 <= f.debug)
                      fprintf (outfile, "processline() State: ##%s##\n", State);
                    if (0 == strcmp(State, "01")) {
                      saddr_sport = proc2ip(saddr_sport, resultsaddr_sport);
                      daddr_dport = proc2ip(daddr_dport, resultdaddr_dport);
                    }
                    break;
    default :
                    pid = 0;
#if defined(__FreeBSD__) || defined(__linux__)
                    /* Proto */ shift(&remainder, 1 /* stop at first space */);
                    /* Recv_Q */ shift(&remainder, 1 /* stop at first space */);
                    /* Send_Q */ shift(&remainder, 1 /* stop at first space */);
                    saddr_sport = shift(&remainder, 1 /* stop at first space */);
                    daddr_dport = shift(&remainder, 1 /* stop at first space */);
#else
                    saddr_sport = shift(&remainder, 1 /* stop at first space */);
                    daddr_dport = shift(&remainder, 1 /* stop at first space */);
                    /* Swind */ shift(&remainder, 1 /* stop at first space */);
                    /* Send_Q */ shift(&remainder, 1 /* stop at first space */);
                    /* Rwind */ shift(&remainder, 1 /* stop at first space */);
                    /* Recv_Q */ shift(&remainder, 1 /* stop at first space */);
#endif
#if defined(__linux__)
                    State = shift(&remainder, 1 /* stop at first space */);
                    { char *p;

                      pid = 0;
                      p = shift(&remainder, 0 /* do not stop at space */);
                      while ('0' <= *p && *p <= '9') {
                        pid = pid * 10 + (*p - '0');
                        p++;
                      }
                    }
#else
                    State = shift(&remainder, 0 /* do not stop at space */);
#endif
  }

  if (4 <= f.debug)
    fprintf (outfile, "processline(): state ##%s##\n", State);

  if (strncmp(State, (LSOFMODE == f.callmode) ? "(ESTABLISHED)" : "ESTABLISHED", sizeof((LSOFMODE == f.callmode) ? "(ESTABLISHED)" : "ESTABLISHED") - 1) == 0 || (CATPROCMODE == f.callmode && 0 == strcmp(State, "01"))) {
    normalize(saddr_sport, f.callmode);
    normalize(daddr_dport, f.callmode);
    for (i = 0; i < lastsocket; i++) {
      if (matching(saddr_sport, socketlist[i].saddr_sport, 0) &&
          matching(daddr_dport, socketlist[i].daddr_dport, 0)) {
        socketlist[i].loop = loop;
        return;
      }
    }
    if (lastsocket == MAXLASTSOCKET)
      myerrx(1, "increase MAXLASTSOCKET");
    socketlist[lastsocket].pid = pid;
    socketlist[lastsocket].saddr_sport = strdup(saddr_sport);
    socketlist[lastsocket].daddr_dport = strdup(daddr_dport);
    socketlist[lastsocket].loop = loop;
    socketlist[lastsocket].start_hour_min = time_hour_min;
    socketlist[lastsocket].match = -1;
    socketlist[lastsocket].state = STATEUNKNOWN;
    lastsocket++;
    if (2 <= f.debug || time_hour_min < record)
      fprintf(outfile, "+ %04d 2400 %ld %s %s\n",
              time_hour_min, (long)pid, saddr_sport, daddr_dport);
  }
}

int
process(void)
{ int e;

  /* remove ended connections (those with a not up to date tag) */
  removeoldconnections();
  /* unmark connections : set to state STATEUNKNOWN */
  resetstate();
  /* alarm on missing connections */
  e = missing();
  /* alarm on established sockets during wrong period */
  running();
  if (f.excluded) {
    markexcluded();
    unknown();
  }
  if (f.itrs && (f.once == 1 || e == 0))
    dumpconfigtoitrs();
  return (e);
}

void
usage(void)
{
  fprintf(stderr, "%s\n%s\n",
      "usage: netchecker [-1DLPcdeginqrsv] [-E excludefilename] [-o outfile]",
      "                  [-p probehost] [-R time] [-t timer] conffilename");
  exit(1);
}

void
terminate(int notused)
{
  sig_terminate = 1;
}

void
initprobe(void)
{ struct hostent *hp;
  struct servent *sserv;
  int i;

  /* probehost is set by readconffile() or by -p option flag */
  for (i = 0; i < lastprobehost; i++) {
    if (f.debug != 0)
      fprintf(outfile, "probehost set to ##%s## with public=%d\n",
          probehostlist[i].probehost, probehostlist[i].public);
    if ((probehostlist[i].sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
      myerrx(1, "socket");

    hp = gethostbyname(probehostlist[i].probehost);
    if (hp == NULL)
      myerrx(1, "probehost: host not found in database");

    sserv = getservbyname("phchaprobe", "udp");
    if (sserv == NULL)
      myerrx(1, "phchaprobe/udp: service not found in database");

    bcopy(hp->h_addr, &(probehostlist[i].name).sin_addr, hp->h_length);
    (probehostlist[i].name).sin_family = AF_INET;
    (probehostlist[i].name).sin_port = sserv->s_port;
  }
}

void
doneprobe(void)
{ int i;

  for (i = 0; i < lastprobehost; i++)
    close(probehostlist[i].sockfd);
}

int
main (int argc, char **argv)
{ char buf[MAXBUFFERSIZE];
  int fildes[2];
  int ch;
  time_t thetime;
  struct tm *ptm;
  char msg[1024];
  pid_t	pid_child;
  FILE *stream;
  int status;
  int e;

  loop = 1;
  lastsocket = 0;
  lastprobehost = 0;
  time_hour_min = 0;
  time_hour_min_sec = 0;
  weekday = atobin("1111111");
  category = strdup("n/a");
  f.debug = 0;
  f.dump_netstat = 0;
  f.guessdot = 0;
  f.itrs = 0;
  f.callmode = NETSTATMODE;
  f.once = 0;
  f.syslog = 0;
  f.displayconfig = 0;
  f.quiet = 0;
  f.regex = 0;
  f.excluded = 0;
  f.numeric = 0;
  excludefilename = NULL;
  outfile = stdout;
  sig_terminate = 0;
  timer = 30;
  record = 0;
  socketsallowed = 100;
  e = 0;

  while ((ch = getopt(argc, argv, "1DE:LR:cdegino:Pp:qrst:v")) != EOF)
    switch(ch) {
    case '1' : f.once++;
               break;
    case 'c' : f.displayconfig++;
               break;
    case 'D' : f.dump_netstat++;
               break;
    case 'd' : f.debug++;
               break;
    case 'E' : excludefilename = optarg;
               break;
    case 'e' : f.excluded++;
               break;
    case 'g' : f.guessdot++;
               break;
    case 'i' : f.itrs++;
               break;
    case 'L' : f.callmode = LSOFMODE;
               break;
    case 'n' : f.numeric = 1;
               break;
    case 'o' : outfile = fopen(optarg, "w");
               if (outfile == NULL)
                 myerrx(1, "incorrect output file name");
               break;
    case 'P' : f.callmode = CATPROCMODE;
               break;
    case 'p' : if (lastprobehost == MAXPROBELIST)
                 myerrx(1, "increase MAXPROBELIST");
               probehostlist[lastprobehost].public = -1;
               probehostlist[lastprobehost].probehost = optarg;
               lastprobehost++;
               break;
    case 'q' : f.quiet++;
               break;
    case 'R' : record = atoi(optarg);
               if (record < 0 || 2400 < record)
                 myerrx(1, "-R value not in [0000..2400]");
               break;
    case 'r' : f.regex = 1;
               break;
    case 's' : f.syslog++;
               break;
    case 't' : timer = atoi(optarg);
               break;
    case 'v' : printf("%s\n", rcsid);
               exit(1);
               break;
    default  : usage();
    }
  argc -= optind;
  argv += optind;

  if ((conffilename = argv[0]) == NULL)
    usage();

  signal(SIGINT, terminate);
  signal(SIGTERM, terminate);

  if (f.debug == 0 && 1 <= f.syslog)
    openlog("netchecker", 0 /*LOG_PID*/, LOG_USER);

  lastexclude = 0;
  readconffile();
  if (f.displayconfig != 0) {
    displayconfig();
    if (f.debug == 0 && 1 <= f.syslog)
      closelog();
    return (0);
  }
  readexcludefile();

  gethostname(hostname, sizeof(hostname));

  initprobe();

  while (! sig_terminate) {
    pipe(fildes);

    switch((pid_child = fork())) {
       case -1 : myerrx(1, "fork failed");
       case 0  : /* child */
                 close(fildes[0]);      /* close the half pipe we don't use */
                 close(STDOUT_FILENO);  /* and stdout */
                 close(STDERR_FILENO);  /* and stderr (mode /bin/dd) */
                 dup2(fildes[1], STDOUT_FILENO); /* stdout -> pipe */
                 switch (f.callmode) {
                   case LSOFMODE :
#if defined(__linux__)
                                 execl("/usr/sbin/lsof", "lsof", ((f.numeric == 1) ? "-sliP" : "-sli"), "TCP", NULL);
#else
                                 execl("/files/misc/bin/lsof", "lsof", ((f.numeric == 1) ? "-sliP" : "-sli"), "TCP", NULL);
#endif
                                 break;
                   case CATPROCMODE :
                                 /* execl("/bin/cat", "cat", "/proc/net/tcp", NULL); */
                                 execl("/bin/dd", "dd", "if=/proc/net/tcp", "bs=128k", NULL);
                                 /* execl("/bin/cat", "cat", "/tmp/1", NULL); */
                                 break;
                   default :
#if defined(__FreeBSD__)
                                 execl("/usr/bin/netstat", "netstat", ((f.numeric == 1) ? "-np" : "-p"), "tcp", NULL);
#elif defined(__linux__)
                                 if (f.numeric == 1)
                                   execl("/bin/netstat", "netstat", "-t", "-p", "--numeric-ports", NULL);
                                 else
                                   execl("/bin/netstat", "netstat", "-t", "-p", NULL);
#else
                                 execl("/usr/bin/netstat", "netstat", ((f.numeric == 1) ? "-nP" : "-P"), "tcp", NULL);
#endif
                 }
                 /* NOTREACHED */
                 myerrx(0, "exec failed");
       default : /* parent */
                 close(fildes[1]);      /* close the half pipe we don't use */
                 if ((stream = fdopen(fildes[0], "r")) == NULL) {
                   kill(pid_child, SIGTERM);
                   exit(1);
                 }

                 sockets = 0;
                 while (fgets(buf, MAXBUFFERSIZE, stream)) {
                   processline(buf);
                   sockets++;
                 }

                 fclose(stream);
                 close(fildes[0]);
                 readpidfile(-1); /* get up to date information */

                 if (waitpid(pid_child, &status, 0) == -1)
                   myerrx(1, "waitpid() failed");

                 thetime = time(NULL);
                 ptm = localtime(&thetime);
                 time_hour_min = ptm->tm_hour*100 + ptm->tm_min;
                 time_hour_min_sec = time_hour_min*100 + ptm->tm_sec;
                 weekday = 1 << (6 - ptm->tm_wday);
                 if (f.debug == 0) {
                   sprintf(msg, ":h=%s:a=N-PROBE:l=%d:", hostname, timer);
                   sendprobemessage(msg, -2);
                 }
                 if (socketsallowed < sockets) {
                   if (1 <= f.syslog)
                     syslog(LOG_ERR, "high number of sockets at %06d: %d/%d ",
                       time_hour_min_sec, sockets, socketsallowed);
                   sprintf(msg, ":h=%s:a=N-NUMSOCKET:l=%d:x=%d/%d:", hostname, timer, sockets, socketsallowed);
                   sendprobemessage(msg, -2);
                 }
                 e = process();
                 dumpall();
                 break;
    }
    if (2 <= f.debug)
      fprintf(outfile, "===hour: %04d==loop: %d=="
              "=( )list (+)add (-)remove===\n", time_hour_min, loop);
    loop++;
    if (1 < f.once) {
      f.once--;
      if (e == 0 || (record != 0 && record <= time_hour_min))
        sig_terminate = 1;
      else
        sleep(3);
    }
    else {
      if (f.once == 1 || (record != 0 && record <= time_hour_min))
        sig_terminate = 1;
      else
        sleep(timer);
    }
  }
  fclose(outfile);
  if (f.debug == 0 && 1 <= f.syslog)
    closelog();

  doneprobe();
  return (0);
}
